#!/usr/bin/env python3
"""
Traffic Monitor
Real-time network traffic monitoring and analysis tool.
Parameters:
    - interface: Network interface to monitor
    - filter: BPF filter for packet capture
    - output: Output file for traffic data
Requires: scapy package and admin/root privileges
"""
from scapy.all import *
import argparse
import logging
from datetime import datetime
import json
import threading
import queue
from collections import defaultdict
import time

class TrafficMonitor:
    def __init__(self, interface, packet_filter=None, max_packets=1000):
        self.interface = interface
        self.packet_filter = packet_filter
        self.max_packets = max_packets
        self.stop_event = threading.Event()
        self.packets = queue.Queue(maxsize=max_packets)
        self.stats = {
            'total_packets': 0,
            'total_bytes': 0,
            'protocols': defaultdict(int),
            'ports': defaultdict(int),
            'ip_pairs': defaultdict(int),
            'packet_sizes': defaultdict(int)
        }
        self.setup_logging()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def packet_callback(self, packet):
        if self.stop_event.is_set():
            return
        
        try:
            # Basic packet info
            packet_info = {
                'timestamp': datetime.now().isoformat(),
                'size': len(packet),
                'summary': packet.summary()
            }

            # Layer-specific information
            if IP in packet:
                packet_info.update({
                    'src_ip': packet[IP].src,
                    'dst_ip': packet[IP].dst,
                    'protocol': packet[IP].proto
                })
                
                # Update IP pair statistics
                ip_pair = f"{packet[IP].src} -> {packet[IP].dst}"
                self.stats['ip_pairs'][ip_pair] += 1

            if TCP in packet:
                packet_info.update({
                    'src_port': packet[TCP].sport,
                    'dst_port': packet[TCP].dport,
                    'flags': packet[TCP].flags
                })
                self.stats['protocols']['TCP'] += 1
                self.stats['ports'][packet[TCP].sport] += 1
                self.stats['ports'][packet[TCP].dport] += 1

            elif UDP in packet:
                packet_info.update({
                    'src_port': packet[UDP].sport,
                    'dst_port': packet[UDP].dport
                })
                self.stats['protocols']['UDP'] += 1
                self.stats['ports'][packet[UDP].sport] += 1
                self.stats['ports'][packet[UDP].dport] += 1

            elif ICMP in packet:
                packet_info.update({
                    'icmp_type': packet[ICMP].type,
                    'icmp_code': packet[ICMP].code
                })
                self.stats['protocols']['ICMP'] += 1

            # Update general statistics
            self.stats['total_packets'] += 1
            self.stats['total_bytes'] += packet_info['size']
            self.stats['packet_sizes'][packet_info['size']] += 1

            # Store packet info
            try:
                self.packets.put(packet_info, block=False)
            except queue.Full:
                # If queue is full, remove oldest packet
                self.packets.get()
                self.packets.put(packet_info)

        except Exception as e:
            self.logger.error(f"Error processing packet: {str(e)}")

    def calculate_bandwidth(self, interval=1):
        """Calculate current bandwidth usage"""
        bytes_start = self.stats['total_bytes']
        time.sleep(interval)
        bytes_end = self.stats['total_bytes']
        bytes_per_sec = (bytes_end - bytes_start) / interval
        return bytes_per_sec

    def start_capture(self):
        try:
            self.logger.info(f"Starting traffic capture on {self.interface}")
            if self.packet_filter:
                self.logger.info(f"Using filter: {self.packet_filter}")
            
            sniff(iface=self.interface, prn=self.packet_callback,
                  filter=self.packet_filter, store=0,
                  stop_filter=lambda _: self.stop_event.is_set())
                  
        except Exception as e:
            self.logger.error(f"Error during packet capture: {str(e)}")

    def save_results(self, output_file):
        packets = []
        while not self.packets.empty():
            packets.append(self.packets.get())

        results = {
            'timestamp': datetime.now().isoformat(),
            'interface': self.interface,
            'filter': self.packet_filter,
            'statistics': {
                'total_packets': self.stats['total_packets'],
                'total_bytes': self.stats['total_bytes'],
                'protocols': dict(self.stats['protocols']),
                'top_ports': dict(sorted(
                    self.stats['ports'].items(),
                    key=lambda x: x[1],
                    reverse=True
                )[:10]),
                'top_ip_pairs': dict(sorted(
                    self.stats['ip_pairs'].items(),
                    key=lambda x: x[1],
                    reverse=True
                )[:10]),
                'packet_size_distribution': dict(self.stats['packet_sizes'])
            },
            'packets': packets
        }

        try:
            with open(output_file, 'w') as f:
                json.dump(results, f, indent=4)
            self.logger.info(f"Results saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving results: {str(e)}")

    def print_stats(self):
        self.logger.info("\nTraffic Statistics:")
        self.logger.info(f"Total Packets: {self.stats['total_packets']}")
        self.logger.info(f"Total Bytes: {self.stats['total_bytes']}")
        
        # Protocol distribution
        self.logger.info("\nProtocol Distribution:")
        total = sum(self.stats['protocols'].values())
        for proto, count in self.stats['protocols'].items():
            percentage = (count / total) * 100 if total > 0 else 0
            self.logger.info(f"{proto}: {count} ({percentage:.1f}%)")
        
        # Top talkers
        self.logger.info("\nTop IP Pairs:")
        for ip_pair, count in sorted(
            self.stats['ip_pairs'].items(),
            key=lambda x: x[1],
            reverse=True
        )[:5]:
            self.logger.info(f"{ip_pair}: {count} packets")

        # Current bandwidth
        bps = self.calculate_bandwidth()
        self.logger.info(f"\nCurrent Bandwidth: {bps/1024:.2f} KB/s")

    def start_monitoring(self, output_file=None):
        capture_thread = threading.Thread(target=self.start_capture)
        capture_thread.daemon = True
        capture_thread.start()
        
        try:
            while True:
                self.print_stats()
                time.sleep(5)
                
        except KeyboardInterrupt:
            self.logger.info("Stopping traffic monitor...")
            self.stop_event.set()
            capture_thread.join()
            
            if output_file:
                self.save_results(output_file)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Traffic Monitor")
    parser.add_argument("interface", help="Network interface to monitor")
    parser.add_argument("--filter", help="BPF filter for packet capture")
    parser.add_argument("--max-packets", type=int, default=1000,
                      help="Maximum number of packets to store")
    parser.add_argument("--output", help="Output file for traffic data")
    args = parser.parse_args()

    monitor = TrafficMonitor(args.interface, args.filter, args.max_packets)
    monitor.start_monitoring(args.output)
